"""
Some I/O utilities to handle `NeXus <https://manual.nexusformat.org/index.html>`_ and
`HDF5 <https://www.hdfgroup.org/solutions/hdf5/>`_ with `h5py <https://www.h5py.org/>`_.
"""

import logging
import os
from contextlib import contextmanager

import h5py
import h5py._hl.selections as selection
from h5py import File as HDF5File  # noqa F401
from silx.io.url import DataUrl
from silx.io.utils import open as hdf5_open

_logger = logging.getLogger(__name__)

__all__ = [
    "get_swmr_mode",
    "check_virtual_sources_exist",
    "from_data_url_to_virtual_source",
    "from_virtual_source_to_data_url",
    "cwd_context",
    "to_target_rel_path",
]

_DEFAULT_SWMR_MODE = None


def get_swmr_mode() -> bool | None:
    """
    Return True if SWMR should be used in the tomotools scope.
    """
    swmr_mode = os.environ.get("TOMOTOOLS_SWMR", _DEFAULT_SWMR_MODE)
    if swmr_mode in (None, "None", "NONE"):
        return None
    else:
        return swmr_mode in (
            True,
            "True",
            "true",
            "TRUE",
            "1",
            1,
        )


def check_virtual_sources_exist(fname, data_path):
    """
    Check that a virtual dataset points to actual data.

    :param fname: HDF5 file path
    :param data_path: Path within the HDF5 file

    :return res: Whether the virtual dataset points to actual data.
    """
    with hdf5_open(fname) as f:
        if data_path not in f:
            _logger.error(f"No dataset {data_path} in file {fname}")
            return False
        dptr = f[data_path]
        if not dptr.is_virtual:
            return True
        for vsource in dptr.virtual_sources():
            vsource_fname = os.path.join(
                os.path.dirname(dptr.file.filename), vsource.file_name
            )
            if not os.path.isfile(vsource_fname):
                _logger.error(f"No such file: {vsource_fname}")
                return False
            elif not check_virtual_sources_exist(vsource_fname, vsource.dset_name):
                _logger.error(f"Error with virtual source {vsource_fname}")
                return False
    return True


def from_data_url_to_virtual_source(url: DataUrl, target_path: str | None) -> tuple:
    """
    Convert a DataUrl to a set (as tuple) of h5py.VirtualSource.

    :param url: URL to be converted to a virtual source. It must target a 2D detector.
    :return: (h5py.VirtualSource, tuple(shape of the virtual source), numpy.dtype: type of the dataset associated with the virtual source)
    """
    if not isinstance(url, DataUrl):
        raise TypeError(
            f"url is expected to be an instance of DataUrl and not {type(url)}"
        )

    with hdf5_open(url.file_path()) as o_h5s:
        original_data_shape = o_h5s[url.data_path()].shape
        data_type = o_h5s[url.data_path()].dtype
        if len(original_data_shape) == 2:
            original_data_shape = (
                1,
                original_data_shape[0],
                original_data_shape[1],
            )

        vs_shape = original_data_shape
        if url.data_slice() is not None:
            vs_shape = (
                url.data_slice().stop - url.data_slice().start,
                original_data_shape[-2],
                original_data_shape[-1],
            )

    if target_path is not None and (
        target_path == url.file_path()
        or os.path.abspath(target_path) == url.file_path()
    ):
        file_path = "."
    else:
        file_path = url.file_path()
    vs = h5py.VirtualSource(file_path, url.data_path(), shape=vs_shape, dtype=data_type)

    if url.data_slice() is not None:
        vs.sel = selection.select(original_data_shape, url.data_slice())
    return vs, vs_shape, data_type


def from_virtual_source_to_data_url(
    vs: h5py.VirtualSource, vs_slice: slice | None = None
) -> DataUrl:
    """
    Convert a h5py.VirtualSource to a DataUrl.

    :param vs: virtual source to be converted to a DataUrl.
    :param vs_slice: virtual source slice. To be provided because cannot be deduced from the public API of ``h5py.VirtualSource``
    :return: resulting URL.
    """
    if not isinstance(vs, h5py.VirtualSource):
        raise TypeError(
            f"vs is expected to be an instance of h5py.VirtualSorce and not {type(vs)}"
        )

    url = DataUrl(
        file_path=vs.path, data_path=vs.name, scheme="silx", data_slice=vs_slice
    )
    return url


@contextmanager
def cwd_context(new_cwd=None):
    """
    Create a context with `new_cwd`.

    On entry update the current working directory to `new_cwd` and reset the previous working directory at exit.

    :param new_cwd: working directory to use in the context.
    """
    try:
        curdir = os.getcwd()
    except Exception as e:
        _logger.error(e)
        curdir = None
    try:
        if new_cwd is not None and os.path.isfile(new_cwd):
            new_cwd = os.path.dirname(new_cwd)
        if new_cwd not in (None, ""):
            os.chdir(new_cwd)
        yield
    finally:
        if curdir is not None:
            os.chdir(curdir)


def to_target_rel_path(file_path: str, target_path: str) -> str:
    """
    Cast `file_path` to a relative path according to `target_path`.
    This is used to deduce the path of an h5py.VirtualSource.

    :param file_path: file path to convert to a relative path.
    :param target_path: reference path used to compute the relative path.
    :return: relative path of `file_path` compared to `target_path`.
    """
    if file_path == target_path or os.path.abspath(file_path) == os.path.abspath(
        target_path
    ):
        return "."
    file_path = os.path.abspath(file_path)
    target_path = os.path.abspath(target_path)
    path = os.path.relpath(file_path, os.path.dirname(target_path))
    if not path.startswith("./"):
        path = "./" + path
    return path



"""
Module for handling an `NXsample <https://manual.nexusformat.org/classes/base_classes/NXsample.html>`_.
"""

import logging
from functools import partial
from operator import is_not

import numpy
import pint
from silx.utils.proxy import docstring

from nxtomo.nxobject.nxobject import NXobject
from nxtomo.nxobject.nxtransformations import NXtransformations
from nxtomo.nxobject.utils.concatenate import concatenate_pint_quantities
from nxtomo.nxobject.utils.decorator import check_dimensionality
from nxtomo.nxobject.utils.ObjectWithPixelSizeMixIn import _ObjectWithPixelSizeMixIn
from nxtomo.paths.nxtomo import get_paths as get_nexus_paths
from nxtomo.utils import get_data, get_quantity

_logger = logging.getLogger(__name__)

__all__ = [
    "NXsample",
]

_ureg = pint.UnitRegistry()
_meter = _ureg.meter
_degree = _ureg.degree


class NXsample(NXobject, _ObjectWithPixelSizeMixIn):
    def __init__(self, node_name="sample", parent: NXobject | None = None) -> None:
        """
        Representation of `NeXus NXsample <https://manual.nexusformat.org/classes/base_classes/NXsample.html>`_.
        Metadata describing the sample.

        :param node_name: name of the sample in the hierarchy.
        :param parent: parent in the NeXus hierarchy.
        """
        NXobject.__init__(self, node_name=node_name, parent=parent)
        self._set_freeze(False)
        _ObjectWithPixelSizeMixIn.__init__(self)

        self._name = None
        self._rotation_angle = None
        self._x_translation: pint.Quantity | None = None
        self._y_translation: pint.Quantity | None = None
        self._z_translation: pint.Quantity | None = None
        self._propagation_distance: pint.Quantity | None = None
        self._transformations = tuple()
        self._set_freeze(True)

    @property
    def name(self) -> str | None:
        """Sample name."""
        return self._name

    @name.setter
    def name(self, name: str | None) -> None:
        if not isinstance(name, (type(None), str)):
            raise TypeError(f"name is expected to be None or str not {type(name)}")
        self._name = name

    @property
    def rotation_angle(self) -> pint.Quantity | None:
        """Sample rotation angle (one value per frame)."""
        return self._rotation_angle

    @rotation_angle.setter
    @check_dimensionality("[]")
    def rotation_angle(self, rotation_angle: None | pint.Quantity):
        self._rotation_angle = rotation_angle

    @property
    def x_translation(self) -> pint.Quantity | None:
        """Sample translation along X. See `modelling at ESRF <https://tomo.gitlab-pages.esrf.fr/ebs-tomo/master/modelization.html>`_ for more information."""
        return self._x_translation

    @x_translation.setter
    @check_dimensionality("[length]")
    def x_translation(self, x_translation: None | pint.Quantity):
        self._x_translation = x_translation

    @property
    def y_translation(self) -> pint.Quantity | None:
        """Sample translation along Y. See `modelling at ESRF <https://tomo.gitlab-pages.esrf.fr/ebs-tomo/master/modelization.html>`_ for more information."""
        return self._y_translation

    @y_translation.setter
    @check_dimensionality("[length]")
    def y_translation(self, y_translation: None | pint.Quantity):
        self._y_translation = y_translation

    @property
    def z_translation(self) -> pint.Quantity | None:
        """Sample translation along Z. See `modelling at ESRF <https://tomo.gitlab-pages.esrf.fr/ebs-tomo/master/modelization.html>`_ for more information."""
        return self._z_translation

    @z_translation.setter
    @check_dimensionality("[length]")
    def z_translation(self, z_translation: None | pint.Quantity):
        self._z_translation = z_translation

    @property
    def propagation_distance(self) -> pint.Quantity | None:
        return self._propagation_distance

    @propagation_distance.setter
    @check_dimensionality("[length]")
    def propagation_distance(self, value: pint.Quantity | None):
        self._propagation_distance = value

    @property
    def transformations(self) -> tuple[NXtransformations]:
        """Sample transformations as `NXtransformations <https://manual.nexusformat.org/classes/base_classes/NXtransformations.html>`_."""
        return self._transformations

    @transformations.setter
    def transformations(self, transformations: tuple[NXtransformations]):
        if not isinstance(transformations, tuple):
            raise TypeError(
                f"transformations expects a tuple. Got {type(transformations)}"
            )
        for transformation in transformations:
            if not isinstance(transformation, NXtransformations):
                raise TypeError(
                    f"transformations should be a tuple of {NXtransformations}. Contains {type(transformation)}"
                )

    @docstring(NXobject)
    def to_nx_dict(
        self,
        nexus_path_version: float | None = None,
        data_path: str | None = None,
    ) -> dict:
        nexus_paths = get_nexus_paths(nexus_path_version)
        nexus_sample_paths = nexus_paths.nx_sample_paths
        x_pixel_size_path = (
            "/".join([self.path, nexus_sample_paths.X_PIXEL_SIZE])
            if nexus_sample_paths.X_PIXEL_SIZE is not None
            else None
        )
        y_pixel_size_path = (
            "/".join([self.path, nexus_sample_paths.Y_PIXEL_SIZE])
            if nexus_sample_paths.Y_PIXEL_SIZE is not None
            else None
        )

        nx_dict = _ObjectWithPixelSizeMixIn.to_nx_dict(
            self,
            x_pixel_size_path=x_pixel_size_path,
            y_pixel_size_path=y_pixel_size_path,
        )

        if self.name is not None:
            path_name = f"{self.path}/{nexus_sample_paths.NAME}"
            nx_dict[path_name] = self.name
        if self.rotation_angle is not None:
            path_rotation_angle = f"{self.path}/{nexus_sample_paths.ROTATION_ANGLE}"
            nx_dict[path_rotation_angle] = self.rotation_angle.to(
                _ureg.degree
            ).magnitude
            nx_dict["@".join([path_rotation_angle, "units"])] = "degree"
        if self.x_translation is not None:
            path_x_translation = f"{self.path}/{nexus_sample_paths.X_TRANSLATION}"
            nx_dict[path_x_translation] = self.x_translation.magnitude
            nx_dict["@".join([path_x_translation, "units"])] = (
                f"{self.x_translation.units:~}"
            )
        if self.y_translation is not None:
            path_y_translation = f"{self.path}/{nexus_sample_paths.Y_TRANSLATION}"
            nx_dict[path_y_translation] = self.y_translation.magnitude
            nx_dict["@".join([path_y_translation, "units"])] = (
                f"{self.y_translation.units:~}"
            )
        if self.z_translation is not None:
            path_z_translation = f"{self.path}/{nexus_sample_paths.Z_TRANSLATION}"
            nx_dict[path_z_translation] = self.z_translation.magnitude
            nx_dict["@".join([path_z_translation, "units"])] = (
                f"{self.z_translation.units:~}"
            )

        if (
            self.propagation_distance is not None
            and nexus_sample_paths.PROPAGATION_DISTANCE is not None
        ):
            path_propagation_distance = (
                f"{self.path}/{nexus_sample_paths.PROPAGATION_DISTANCE}"
            )
            nx_dict[path_propagation_distance] = self.propagation_distance
            nx_dict["@".join([path_propagation_distance, "units"])] = (
                f"{self.propagation_distance.units:~}"
            )

        if nx_dict != {}:
            nx_dict[f"{self.path}@NX_class"] = "NXsample"
        return nx_dict

    def _load(self, file_path: str, data_path: str, nexus_version: float) -> NXobject:
        """
        Create and load an NXsample from data on disk.
        """
        nexus_paths = get_nexus_paths(nexus_version)
        nexus_sample_paths = nexus_paths.nx_sample_paths

        _ObjectWithPixelSizeMixIn._load(
            self,
            file_path=file_path,
            x_pixel_size_path=(
                "/".join([data_path, nexus_sample_paths.X_PIXEL_SIZE])
                if nexus_sample_paths.X_PIXEL_SIZE
                else None
            ),
            y_pixel_size_path=(
                "/".join([data_path, nexus_sample_paths.Y_PIXEL_SIZE])
                if nexus_sample_paths.Y_PIXEL_SIZE
                else None
            ),
        )

        self.name = get_data(
            file_path=file_path,
            data_path="/".join([data_path, nexus_sample_paths.NAME]),
        )
        self.rotation_angle = get_quantity(
            file_path=file_path,
            data_path="/".join([data_path, nexus_sample_paths.ROTATION_ANGLE]),
            default_unit=_degree,
        )
        self.x_translation = get_quantity(
            file_path=file_path,
            data_path="/".join([data_path, nexus_sample_paths.X_TRANSLATION]),
            default_unit=_meter,
        )
        self.y_translation = get_quantity(
            file_path=file_path,
            data_path="/".join([data_path, nexus_sample_paths.Y_TRANSLATION]),
            default_unit=_meter,
        )
        self.z_translation = get_quantity(
            file_path=file_path,
            data_path="/".join([data_path, nexus_sample_paths.Z_TRANSLATION]),
            default_unit=_meter,
        )
        if nexus_sample_paths.PROPAGATION_DISTANCE is not None:
            self.propagation_distance = get_quantity(
                file_path=file_path,
                data_path="/".join(
                    [data_path, nexus_sample_paths.PROPAGATION_DISTANCE]
                ),
                default_unit=_meter,
            )

    @staticmethod
    @docstring(NXobject)
    def concatenate(nx_objects: tuple, node_name="sample"):
        nx_objects = tuple(filter(partial(is_not, None), nx_objects))
        # filter None obj
        if len(nx_objects) == 0:
            return None
        # warning: later we make the assumption that nx_objects contains at least one element
        for nx_obj in nx_objects:
            if not isinstance(nx_obj, NXsample):
                raise TypeError("Cannot concatenate non-NXsample object")

        nx_sample = NXsample(node_name)
        _logger.info(f"sample name {nx_objects[0].name} will be picked")
        nx_sample.name = nx_objects[0].name
        _ObjectWithPixelSizeMixIn.concatenate(nx_sample, nx_objects=nx_objects)

        propagation_distance = nx_objects[0].propagation_distance
        if propagation_distance is not None:
            _logger.info(
                f"sample propagation distance {propagation_distance} will be picked"
            )
            nx_sample.propagation_distance = nx_objects[0].propagation_distance
            nx_sample.propagation_distance.unit = nx_objects[
                0
            ].propagation_distance.units

        rotation_angles = [
            nx_obj.rotation_angle
            for nx_obj in nx_objects
            if nx_obj.rotation_angle is not None
        ]
        if rotation_angles:
            nx_sample.rotation_angle = numpy.concatenate(rotation_angles)

        def get_quantities(attr_name):
            values = [
                getattr(nx_obj, attr_name)
                for nx_obj in nx_objects
                if getattr(nx_obj, attr_name) is not None
            ]
            return values

        # Translation attributes
        nx_sample.x_translation = concatenate_pint_quantities(
            get_quantities("x_translation")
        )
        nx_sample.y_translation = concatenate_pint_quantities(
            get_quantities("y_translation")
        )
        nx_sample.z_translation = concatenate_pint_quantities(
            get_quantities("z_translation")
        )

        _ObjectWithPixelSizeMixIn.concatenate(
            output_nx_object=nx_sample, nx_objects=nx_objects
        )

        return nx_sample



"""NeXus paths used to define an `NXsample <https://manual.nexusformat.org/classes/base_classes/NXsample.html>`_."""

from . import nxtransformations


class NEXUS_SAMPLE_PATH:
    NAME = "sample_name"

    ROTATION_ANGLE = "rotation_angle"

    X_TRANSLATION = "x_translation"

    Y_TRANSLATION = "y_translation"

    Z_TRANSLATION = "z_translation"

    NX_TRANSFORMATIONS = None

    NX_TRANSFORMATIONS_PATHS = None

    PROPAGATION_DISTANCE = None

    X_PIXEL_SIZE = None

    Y_PIXEL_SIZE = None


class NEXUS_SAMPLE_PATH_V_1_0(NEXUS_SAMPLE_PATH):
    pass


class NEXUS_SAMPLE_PATH_V_1_1(NEXUS_SAMPLE_PATH_V_1_0):
    NAME = "name"


class NEXUS_SAMPLE_PATH_V_1_2(NEXUS_SAMPLE_PATH_V_1_1):
    pass


class NEXUS_SAMPLE_PATH_V_1_3(NEXUS_SAMPLE_PATH_V_1_2):
    NX_TRANSFORMATIONS = "transformations"

    NX_TRANSFORMATIONS_PATHS = nxtransformations.NEXUS_TRANSFORMATIONS_PATH_V_1_3


class NEXUS_SAMPLE_PATH_V_1_4(NEXUS_SAMPLE_PATH_V_1_3):
    pass


class NEXUS_SAMPLE_PATH_V_1_5(NEXUS_SAMPLE_PATH_V_1_4):
    PROPAGATION_DISTANCE = "propagation_distance"

    X_PIXEL_SIZE = "x_pixel_size"

    Y_PIXEL_SIZE = "y_pixel_size"



import os

import h5py
import numpy
import pint
import pytest
from silx.io.url import DataUrl

from nxtomo.application.nxtomo import NXtomo
from nxtomo.utils.NXtomoSplitter import NXtomoSplitter

ureg = pint.get_application_registry()
degree = ureg.degree
meter = ureg.meter


@pytest.fixture
def nxtomo_to_split(tmp_path) -> tuple[NXtomo, numpy.ndarray]:
    """
    A simple NXtomo for test that contains VDS composed of n Virtual source
    Returns NXtomo, expected_data as numpy array
    """
    n_file = 5
    n_frame_per_file = 20
    layout = h5py.VirtualLayout(
        shape=(n_file * n_frame_per_file, 100, 100), dtype=float
    )
    for i_file in range(n_file):
        file_path = os.path.join(tmp_path, f"file{i_file}.hdf5")
        data_path = f"path_to_dataset_{i_file}"
        with h5py.File(file_path, mode="w") as h5f:
            if i_file == 0:
                data = numpy.ones([n_frame_per_file, 100, 100])
            elif i_file == n_file - 1:
                data = numpy.ones([n_frame_per_file, 100, 100]) * 2
            else:
                start = i_file * 1000.0
                stop = i_file * 1000.0 + (n_frame_per_file * 100 * 100)
                data = numpy.arange(start, stop).reshape(n_frame_per_file, 100, 100)
            h5f[data_path] = data
            vs = h5py.VirtualSource(h5f[data_path])
            layout[i_file * n_frame_per_file : (i_file + 1) * n_frame_per_file] = vs

    master_file = os.path.join(tmp_path, "master_file.hdf5")
    with h5py.File(master_file, mode="w") as h5f:
        h5f.create_virtual_dataset("data", layout)

    nx_tomo = NXtomo()
    with h5py.File(master_file, mode="r") as h5f:
        original_data = h5f["data"][()]
        vs_ = []
        for vs_info in h5f["data"].virtual_sources():
            vs_.append(
                h5py.VirtualSource(
                    vs_info.file_name,
                    vs_info.dset_name,
                    shape=(n_frame_per_file, 100, 100),
                )
            )

        nx_tomo.instrument.detector.data = vs_
    return nx_tomo, original_data


def test_NXtomoSplitter_get_invalid_datasets():
    """Test the NXtomoSplitter `get_invalid_datasets` function."""
    nx_tomo = NXtomo()
    n_frames = 10
    nx_tomo.instrument.detector.data = numpy.random.random(100 * 100 * 10).reshape(
        [n_frames, 100, 100]
    )

    splitter = NXtomoSplitter(nx_tomo=nx_tomo)

    assert len(splitter.get_invalid_datasets()) == 0

    # test rotation angle
    nx_tomo.sample.rotation_angle = [12, 13] * degree
    assert len(splitter.get_invalid_datasets()) == 1
    nx_tomo.sample.rotation_angle = [0] * n_frames * degree
    assert len(splitter.get_invalid_datasets()) == 0

    # test image_key_control
    nx_tomo.instrument.detector.image_key_control = [0]
    assert len(splitter.get_invalid_datasets()) == 1
    nx_tomo.instrument.detector.image_key_control = [0] * n_frames
    assert len(splitter.get_invalid_datasets()) == 0

    # test x_translation
    nx_tomo.sample.x_translation = [0] * meter
    assert len(splitter.get_invalid_datasets()) == 1
    nx_tomo.sample.x_translation = [0] * n_frames * meter
    assert len(splitter.get_invalid_datasets()) == 0

    # test y_translation
    nx_tomo.sample.y_translation = [0] * meter
    assert len(splitter.get_invalid_datasets()) == 1
    nx_tomo.sample.y_translation = [0] * n_frames * meter
    assert len(splitter.get_invalid_datasets()) == 0

    # test z_translation
    nx_tomo.sample.z_translation = [0] * meter
    assert len(splitter.get_invalid_datasets()) == 1
    nx_tomo.sample.z_translation = [0] * n_frames * meter
    assert len(splitter.get_invalid_datasets()) == 0


def test_spliter_raw_data():
    """Test the splitter on a simple non-virtual h5py dataset."""
    nx_tomo = NXtomo()
    n_frames = 20
    nx_tomo.instrument.detector.data = numpy.random.random(
        100 * 100 * n_frames
    ).reshape([n_frames, 100, 100])
    nx_tomo.sample.rotation_angle = [0, 12] * degree
    nx_tomo.instrument.detector.sequence_number = numpy.linspace(
        0, n_frames, n_frames, dtype=numpy.uint32
    )

    splitter = NXtomoSplitter(nx_tomo=nx_tomo)
    # check incoherent number of rotation
    with pytest.raises(ValueError):
        splitter.split(data_slice=slice(0, 100, 1), nb_part=2)
    nx_tomo.sample.rotation_angle = (
        numpy.linspace(0, 180, num=n_frames, endpoint=False) * degree
    )
    # check slice nb_part < 0
    with pytest.raises(ValueError):
        splitter.split(data_slice=slice(0, 100, 1), nb_part=-1)
    # check slice step != 1
    with pytest.raises(ValueError):
        splitter.split(data_slice=slice(0, 100, 2), nb_part=2)
    # check incoherent number of frames
    with pytest.raises(ValueError):
        splitter.split(data_slice=slice(0, 99, 2), nb_part=2)

    # check x translation
    nx_tomo.sample.x_translation = [0, 12] * meter
    with pytest.raises(ValueError):
        splitter.split(data_slice=slice(0, 100, 1), nb_part=2)
    nx_tomo.sample.x_translation = numpy.random.random(n_frames) * meter
    nx_tomo.sample.y_translation = numpy.random.random(n_frames) * meter
    nx_tomo.sample.z_translation = numpy.random.random(n_frames) * meter
    # check image key
    nx_tomo.instrument.detector.image_key_control = [0, 2]
    with pytest.raises(ValueError):
        splitter.split(data_slice=slice(0, 100, 1), nb_part=2)
    nx_tomo.instrument.detector.image_key_control = [
        numpy.random.randint(low=-1, high=2) for i in range(n_frames)
    ]

    assert splitter.split(data_slice=slice(0, 100, 1), nb_part=1) == [
        nx_tomo,
    ]

    # check error if request to split a region bigger that the one (100 vs n_frames)
    with pytest.raises(ValueError):
        splitted_nx_tomo = splitter.split(data_slice=slice(0, 100, 1), nb_part=2)

    splitted_nx_tomo = splitter.split(data_slice=slice(0, 20, 1), nb_part=2)
    assert len(splitted_nx_tomo) == 2
    s_nx_tomo_1, s_nx_tomo_2 = splitted_nx_tomo
    # chek rotation_angle
    numpy.testing.assert_array_equal(
        s_nx_tomo_1.sample.rotation_angle.magnitude,
        nx_tomo.sample.rotation_angle.magnitude[0 : n_frames // 2],
    )
    assert str(s_nx_tomo_1.sample.rotation_angle.units) == str(
        nx_tomo.sample.rotation_angle.units
    )
    numpy.testing.assert_array_equal(
        s_nx_tomo_2.sample.rotation_angle.magnitude,
        nx_tomo.sample.rotation_angle.magnitude[n_frames // 2 :],
    )
    assert str(s_nx_tomo_2.sample.rotation_angle.units) == str(
        nx_tomo.sample.rotation_angle.units
    )
    # check image key and image key
    numpy.testing.assert_array_equal(
        s_nx_tomo_1.instrument.detector.image_key_control,
        nx_tomo.instrument.detector.image_key_control[0 : n_frames // 2],
    )
    numpy.testing.assert_array_equal(
        s_nx_tomo_2.instrument.detector.image_key_control,
        nx_tomo.instrument.detector.image_key_control[n_frames // 2 :],
    )
    # check sequence_number
    numpy.testing.assert_array_equal(
        s_nx_tomo_1.instrument.detector.sequence_number,
        numpy.linspace(0, n_frames // 2 - 1, n_frames // 2, dtype=numpy.uint32),
    )
    numpy.testing.assert_array_equal(
        s_nx_tomo_2.instrument.detector.sequence_number,
        numpy.linspace(n_frames // 2, n_frames, n_frames // 2, dtype=numpy.uint32),
    )

    # chek x translation
    numpy.testing.assert_array_equal(
        s_nx_tomo_1.sample.x_translation.magnitude,
        nx_tomo.sample.x_translation.magnitude[0 : n_frames // 2],
    )
    assert str(s_nx_tomo_1.sample.x_translation.units) == str(
        nx_tomo.sample.x_translation.units
    )
    numpy.testing.assert_array_equal(
        s_nx_tomo_2.sample.x_translation.magnitude,
        nx_tomo.sample.x_translation.magnitude[n_frames // 2 :],
    )
    assert str(s_nx_tomo_2.sample.x_translation.units) == str(
        nx_tomo.sample.x_translation.units
    )

    # chek y translation
    numpy.testing.assert_array_equal(
        s_nx_tomo_1.sample.y_translation.magnitude,
        nx_tomo.sample.y_translation.magnitude[0 : n_frames // 2],
    )
    assert str(s_nx_tomo_1.sample.y_translation.units) == str(
        nx_tomo.sample.y_translation.units
    )
    numpy.testing.assert_array_equal(
        s_nx_tomo_2.sample.y_translation.magnitude,
        nx_tomo.sample.y_translation.magnitude[n_frames // 2 :],
    )
    assert str(s_nx_tomo_2.sample.y_translation.units) == str(
        nx_tomo.sample.y_translation.units
    )
    # chek z translation
    numpy.testing.assert_array_equal(
        s_nx_tomo_1.sample.z_translation.magnitude,
        nx_tomo.sample.z_translation.magnitude[0 : n_frames // 2],
    )
    assert str(s_nx_tomo_1.sample.z_translation.units) == str(
        nx_tomo.sample.z_translation.units
    )
    numpy.testing.assert_array_equal(
        s_nx_tomo_2.sample.z_translation.magnitude,
        nx_tomo.sample.z_translation.magnitude[n_frames // 2 :],
    )
    assert str(s_nx_tomo_2.sample.z_translation.units) == str(
        nx_tomo.sample.z_translation.units
    )
    # check detector data
    numpy.testing.assert_array_equal(
        s_nx_tomo_1.instrument.detector.data,
        nx_tomo.instrument.detector.data[0 : n_frames // 2],
    )
    numpy.testing.assert_array_equal(
        s_nx_tomo_2.instrument.detector.data,
        nx_tomo.instrument.detector.data[n_frames // 2 :],
    )


def test_splitter_tomo_n_handles_remainder_and_single_frame():
    """Ensure splitting by tomo_n keeps leftover frames and single-frame chunks."""
    nx_tomo = NXtomo()
    n_frames = 5
    nx_tomo.instrument.detector.data = numpy.random.random(4 * 4 * n_frames).reshape(
        [n_frames, 4, 4]
    )
    nx_tomo.instrument.detector.image_key_control = [0] * n_frames
    nx_tomo.sample.rotation_angle = numpy.arange(n_frames) * degree
    nx_tomo.sample.x_translation = numpy.zeros(n_frames) * meter
    nx_tomo.sample.y_translation = numpy.zeros(n_frames) * meter
    nx_tomo.sample.z_translation = numpy.zeros(n_frames) * meter

    splitter = NXtomoSplitter(nx_tomo=nx_tomo)

    parts = splitter.split(
        data_slice=slice(0, n_frames, 1),
        nb_part=None,
        tomo_n=2,
    )
    assert [part.instrument.detector.data.shape[0] for part in parts] == [2, 2, 1]

    single_frame_parts = splitter.split(
        data_slice=slice(0, n_frames, 1),
        nb_part=None,
        tomo_n=1,
    )
    assert len(single_frame_parts) == n_frames
    assert all(
        part.instrument.detector.data.shape[0] == 1 for part in single_frame_parts
    )


def test_spliter_virtual_sources_1():
    """
    Test the splitter on a simulated h5py virtual dataset composed of two virtual sources.
    Both resulting NXtomo instances must reference the same virtual sources.
    Rotation_angle, [W]_translation, and image_key datasets are handled as NumPy arrays
    that do not point to external resources, so only `detector.data` is tested here.
    """
    nx_tomo = NXtomo()
    nx_tomo.instrument.detector.data = [
        h5py.VirtualSource("path_to_dataset_1", name="dataset_1", shape=[10, 100, 100]),
        h5py.VirtualSource("path_to_dataset_2", name="dataset_2", shape=[10, 100, 100]),
    ]
    splitter = NXtomoSplitter(nx_tomo=nx_tomo)
    splitted_nx_tomo = splitter.split(data_slice=slice(0, 20, 1), nb_part=2)
    assert len(splitted_nx_tomo) == 2
    s_nx_tomo_1, s_nx_tomo_2 = splitted_nx_tomo
    det_dataset_1 = s_nx_tomo_1.instrument.detector.data
    det_dataset_2 = s_nx_tomo_2.instrument.detector.data
    assert len(det_dataset_1) == 1
    assert len(det_dataset_2) == 1

    det_dataset_vs1 = det_dataset_1[0]
    det_dataset_vs2 = det_dataset_2[0]
    assert isinstance(det_dataset_vs1, h5py.VirtualSource)
    assert det_dataset_vs1.path == "path_to_dataset_1"
    assert det_dataset_vs1.shape == (10, 100, 100)

    assert isinstance(det_dataset_vs2, h5py.VirtualSource)
    assert det_dataset_vs2.path == "path_to_dataset_2"
    assert det_dataset_vs2.shape == (10, 100, 100)


def test_spliter_virtual_sources_2():
    """
    Test the splitter on an h5py virtual dataset composed of a single virtual source.
    It must split this source into two VirtualSource objects.
    Rotation_angle, [W]_translation, and image_key datasets are handled as NumPy arrays
    that do not point to external resources, so only `detector.data` is tested here.
    """
    nx_tomo = NXtomo()
    nx_tomo.instrument.detector.data = [
        h5py.VirtualSource(
            "path_to_dataset", name="path_to_dataset", shape=[20, 100, 100]
        ),
    ]
    splitter = NXtomoSplitter(nx_tomo=nx_tomo)
    splitted_nx_tomo = splitter.split(data_slice=slice(0, 20, 1), nb_part=2)
    assert len(splitted_nx_tomo) == 2

    splitted_nx_tomo = splitter.split(data_slice=slice(0, 20, 1), nb_part=4)
    assert len(splitted_nx_tomo) == 4


def test_spliter_virtual_sources_3(nxtomo_to_split: str, tmp_path):
    """
    Test the splitter on a concrete h5py virtual dataset.
    Rotation_angle, [W]_translation, and image_key datasets are handled as NumPy arrays
    that do not point to external resources, so only `detector.data` is tested here.
    """
    n_file = 5
    n_frame_per_file = 20

    nx_tomo, original_data = nxtomo_to_split

    output_dir = tmp_path / "output"
    output_dir.mkdir()

    splitter = NXtomoSplitter(nx_tomo=nx_tomo)
    data_slice = slice(10, n_frame_per_file * n_file - 10, 1)
    splitted_nx_tomo = splitter.split(data_slice=data_slice, nb_part=2)
    assert len(splitted_nx_tomo) == 2
    # check the two dataset created
    s_nx_tomo_1, s_nx_tomo_2 = splitted_nx_tomo
    output_file_1 = os.path.join(output_dir, "output_file_1.nx")

    # data must contains a common section between the two nxtomo: the first 10 and last 10 frames
    # then the rest must be splitted between the two NXtomo
    assert len(s_nx_tomo_1.instrument.detector.data) == 5
    assert s_nx_tomo_1.instrument.detector.data[0].shape[0] == 10
    assert s_nx_tomo_1.instrument.detector.data[1].shape[0] == 10
    assert s_nx_tomo_1.instrument.detector.data[2].shape[0] == 20
    assert s_nx_tomo_1.instrument.detector.data[3].shape[0] == 10
    assert s_nx_tomo_1.instrument.detector.data[4].shape[0] == 10
    s_nx_tomo_1.save(output_file_1, "entry0000")

    output_file_2 = os.path.join(output_dir, "output_file_2.nx")
    assert len(s_nx_tomo_2.instrument.detector.data) == 5
    assert s_nx_tomo_2.instrument.detector.data[0].shape[0] == 10
    assert s_nx_tomo_2.instrument.detector.data[1].shape[0] == 10
    assert s_nx_tomo_2.instrument.detector.data[2].shape[0] == 20
    assert s_nx_tomo_2.instrument.detector.data[3].shape[0] == 10
    assert s_nx_tomo_2.instrument.detector.data[4].shape[0] == 10
    s_nx_tomo_2.save(output_file_2, "entry0000")

    # check final datasets are correctly formed
    with h5py.File(output_file_1, mode="r") as h5f:
        nx_1_data = h5f["entry0000/instrument/detector/data"][()]
    assert nx_1_data.shape[0] == 60

    # check final datasets are correctly formed
    with h5py.File(output_file_2, mode="r") as h5f:
        nx_2_data = h5f["entry0000/instrument/detector/data"][()]
    assert nx_2_data.shape[0] == 60

    # first 10 frames (common between the three nxtomo)
    numpy.testing.assert_array_equal(
        nx_1_data[0:10],
        nx_2_data[0:10],
    )
    numpy.testing.assert_array_equal(
        nx_1_data[0:10],
        original_data[0:10],
    )

    # last 10 frames (common between the three nxtomo)
    numpy.testing.assert_array_equal(
        nx_1_data[-10:],
        nx_2_data[-10:],
    )
    numpy.testing.assert_array_equal(
        nx_1_data[-10:],
        original_data[-10:],
    )

    # test nx_1_data unique region
    numpy.testing.assert_array_equal(
        nx_1_data[10:50],
        original_data[10:50],
    )

    # test nx_2_data unique region
    numpy.testing.assert_array_equal(
        nx_2_data[10:50],
        original_data[50:90],
    )


def test_spliter_data_url(tmp_path):
    """
    Test the splitter on a list of DataUrl objects.
    Rotation_angle, [W]_translation, and image_key datasets are handled as NumPy arrays
    that do not point to external resources, so only `detector.data` is tested here.
    """
    urls = []
    n_frame_per_file = 20
    n_file = 5
    original_data = []
    for i_file in range(n_file):
        file_path = os.path.join(tmp_path, f"file{i_file}.hdf5")
        data_path = f"path_to_dataset_{i_file}"
        with h5py.File(file_path, mode="w") as h5f:
            if i_file == 0:
                data = numpy.ones([n_frame_per_file, 100, 100])
            elif i_file == n_file - 1:
                data = numpy.ones([n_frame_per_file, 100, 100]) * 2
            else:
                start = i_file * 1000.0
                stop = i_file * 1000.0 + (n_frame_per_file * 100 * 100)
                data = numpy.arange(start, stop).reshape(n_frame_per_file, 100, 100)
            h5f[data_path] = data
            original_data.append(data)

        urls.append(
            DataUrl(
                file_path=file_path,
                data_path=data_path,
                scheme="silx",
            )
        )

    original_data = numpy.concatenate(original_data)

    nx_tomo = NXtomo()
    nx_tomo.instrument.detector.data = urls

    splitter = NXtomoSplitter(nx_tomo=nx_tomo)
    data_slice = slice(10, n_frame_per_file * n_file - 10, 1)
    data_slice = slice(10, n_frame_per_file * n_file - 10, 1)
    splitted_nx_tomo = splitter.split(data_slice=data_slice, nb_part=2)
    assert len(splitted_nx_tomo) == 2
    # check the two dataset created
    s_nx_tomo_1, s_nx_tomo_2 = splitted_nx_tomo
    output_file_1 = os.path.join(tmp_path, "output_file_1.nx")

    # data must contains a common section between the two nxtomo: the first 10 and last 10 frames
    # then the rest must be splitted between the two NXtomo
    def n_elmt(slice_):
        return slice_.stop - slice_.start

    assert len(s_nx_tomo_1.instrument.detector.data) == 5
    assert n_elmt(s_nx_tomo_1.instrument.detector.data[0].data_slice()) == 10
    assert n_elmt(s_nx_tomo_1.instrument.detector.data[1].data_slice()) == 10
    assert n_elmt(s_nx_tomo_1.instrument.detector.data[2].data_slice()) == 20
    assert n_elmt(s_nx_tomo_1.instrument.detector.data[3].data_slice()) == 10
    assert n_elmt(s_nx_tomo_1.instrument.detector.data[4].data_slice()) == 10
    s_nx_tomo_1.save(output_file_1, "entry0000")

    output_file_2 = os.path.join(tmp_path, "output_file_2.nx")
    assert len(s_nx_tomo_2.instrument.detector.data) == 5
    assert n_elmt(s_nx_tomo_2.instrument.detector.data[0].data_slice()) == 10
    assert n_elmt(s_nx_tomo_2.instrument.detector.data[1].data_slice()) == 10
    assert n_elmt(s_nx_tomo_2.instrument.detector.data[2].data_slice()) == 20
    assert n_elmt(s_nx_tomo_2.instrument.detector.data[3].data_slice()) == 10
    assert n_elmt(s_nx_tomo_2.instrument.detector.data[4].data_slice()) == 10
    s_nx_tomo_2.save(output_file_2, "entry0000")

    # check final datasets are correctly formed
    with h5py.File(output_file_1, mode="r") as h5f:
        nx_1_data = h5f["entry0000/instrument/detector/data"][()]
    assert nx_1_data.shape[0] == 60

    # check final datasets are correctly formed
    with h5py.File(output_file_2, mode="r") as h5f:
        nx_2_data = h5f["entry0000/instrument/detector/data"][()]
    assert nx_2_data.shape[0] == 60

    # first 10 frames (common between the three nxtomo)
    numpy.testing.assert_array_equal(
        nx_1_data[0:10],
        nx_2_data[0:10],
    )
    numpy.testing.assert_array_equal(
        nx_1_data[0:10],
        original_data[0:10],
    )

    # last 10 frames (common between the three nxtomo)
    numpy.testing.assert_array_equal(
        nx_1_data[-10:],
        nx_2_data[-10:],
    )
    numpy.testing.assert_array_equal(
        nx_1_data[-10:],
        original_data[-10:],
    )

    # test nx_1_data unique region
    numpy.testing.assert_array_equal(
        nx_1_data[10:50],
        original_data[10:50],
    )

    # test nx_2_data unique region
    numpy.testing.assert_array_equal(
        nx_2_data[10:50],
        original_data[50:90],
    )


def test_spliter_missing_projections(tmp_path):
    """
    If some projections are missing and `nb_turn` cannot be used, fall back to `tomo_n`.
    """
    urls = []
    n_frame_per_file = 20
    n_file = 5
    original_data = []
    for i_file in range(n_file):
        file_path = os.path.join(tmp_path, f"file{i_file}.hdf5")
        data_path = f"path_to_dataset_{i_file}"
        with h5py.File(file_path, mode="w") as h5f:
            if i_file == 0:
                data = numpy.ones([n_frame_per_file, 100, 100])
            elif i_file == n_file - 1:
                data = numpy.ones([n_frame_per_file, 100, 100]) * 2
            else:
                start = i_file * 1000.0
                stop = i_file * 1000.0 + (n_frame_per_file * 100 * 100)
                data = numpy.arange(start, stop).reshape(n_frame_per_file, 100, 100)
            h5f[data_path] = data
            original_data.append(data)

        urls.append(
            DataUrl(
                file_path=file_path,
                data_path=data_path,
                scheme="silx",
            )
        )

    original_data = numpy.concatenate(original_data)

    nx_tomo = NXtomo()
    nx_tomo.instrument.detector.data = urls

    splitter = NXtomoSplitter(nx_tomo=nx_tomo)
    data_slice = slice(0, 100, 1)
    data_slice = slice(0, 100, 1)
    splitted_nx_tomo = splitter.split(data_slice=data_slice, nb_part=2)
    assert len(splitted_nx_tomo) == 2

    splitted_nx_tomo = splitter.split(data_slice=data_slice, nb_part=None, tomo_n=20)
    assert len(splitted_nx_tomo) == 5

    splitted_nx_tomo = splitter.split(data_slice=data_slice, nb_part=None, tomo_n=40)
    assert len(splitted_nx_tomo) == 3

    splitted_nx_tomo = splitter.split(data_slice=data_slice, nb_part=None, tomo_n=65)
    assert len(splitted_nx_tomo) == 2


@pytest.mark.parametrize(
    "detector_as", ("as_virtual_source", "as_data_url", "as_numpy_array")
)
def test_editing_splitted_nxtomo(nxtomo_to_split: str, detector_as: str, tmp_path):
    """
    Test editing NXtomo that contains a subset of a dataset in a VDS.
    Related to https://gitlab.esrf.fr/tomotools/nxtomo/-/issues/32
    """
    nx_tomo, original_data = nxtomo_to_split

    output_dir = tmp_path / "edition"
    output_dir.mkdir()

    splitter = NXtomoSplitter(nx_tomo=nx_tomo)
    splitted_nx_tomos = splitter.split(
        tomo_n=10, nb_part=None, data_slice=slice(0, 20 * 5, 1)
    )
    # 5 files containing each 20 frames splitted with tomo_n == 10 => 10 nxtomo
    assert len(splitted_nx_tomos) == 10

    for i, nx_tomo in enumerate(splitted_nx_tomos):
        output_file = os.path.join(output_dir, f"splitted_{i}.nx")
        data_path = f"entry{'i'.zfill(4)}"
        nx_tomo.save(output_file, data_path)
        # test loading and saving with the specific 'detector_data_as' attribute
        loaded_nx_tomo = NXtomo().load(
            file_path=output_file,
            data_path=data_path,
            detector_data_as=detector_as,
        )
        loaded_nx_tomo.save(output_file, data_path, overwrite=True)

        # load as numpy array for testing
        nx_tomo_check = NXtomo().load(
            file_path=output_file,
            data_path=data_path,
            detector_data_as="as_numpy_array",
        )
        detector_frames = nx_tomo_check.instrument.detector.data
        assert detector_frames.shape == (10, 100, 100)
        start_slice, end_slice = i * 10, (i + 1) * 10
        numpy.testing.assert_almost_equal(
            detector_frames, original_data[start_slice:end_slice]
        )



"""Define NXtomo application and related functions and classes"""

import logging
import os
from copy import deepcopy
from datetime import datetime
from functools import partial
from operator import is_not

import h5py
import numpy
import pint
from silx.io.url import DataUrl
from silx.io.utils import open as hdf5_open
from silx.utils.proxy import docstring

from nxtomo.geometry._CoordinateSystem import CoordinateSystem
from nxtomo.nxobject.nxdetector import ImageKey
from nxtomo.nxobject.nxinstrument import NXinstrument
from nxtomo.nxobject.nxmonitor import NXmonitor
from nxtomo.nxobject.nxobject import NXobject
from nxtomo.nxobject.nxsample import NXsample
from nxtomo.nxobject.utils.decorator import check_dimensionality
from nxtomo.paths.nxtomo import LATEST_VERSION as LATEST_NXTOMO_VERSION
from nxtomo.paths.nxtomo import get_paths as get_nexus_paths
from nxtomo.utils import get_data, get_quantity

_logger = logging.getLogger(__name__)


__all__ = ["NXtomo", "copy_nxtomo_file"]

_ureg = pint.UnitRegistry()


class NXtomo(NXobject):
    """
    Class defining an NXtomo.
    Its primary goal is to save data to disk.

    :param node_name: node_name is used by the NXobject parent to order children when dumping it to file.
        As NXtomo is expected to be the highest object in the hierarchy, node_name will only be used for saving if no `data_path` is provided when calling the `save` function.
    :param parent: parent of this NXobject. Most likely None for NXtomo.
    """

    def __init__(self, parent: NXobject | None = None) -> None:
        super().__init__(node_name="", parent=parent)
        self._set_freeze(False)
        self._coordinate_system: CoordinateSystem | None = CoordinateSystem.McStas
        self._start_time = None
        self._end_time = None
        self._instrument = NXinstrument(node_name="instrument", parent=self)
        self._sample = NXsample(node_name="sample", parent=self)
        self._control = NXmonitor(node_name="control", parent=self)
        self._group_size = None
        self._bliss_original_files = None  # warning: output will be different if set to None (dataset not exported) or an empty tuple (exported but empty)
        self._energy: pint.Quantity | None = None
        self._title = None
        self._set_freeze(True)

    @property
    def start_time(self) -> datetime | str | None:
        return self._start_time

    @start_time.setter
    def start_time(self, start_time: datetime | str | None):
        if not isinstance(start_time, (type(None), datetime, str)):
            raise TypeError(
                f"start_time is expected ot be an instance of datetime or None. Not {type(start_time)}"
            )
        self._start_time = start_time

    @property
    def end_time(self) -> datetime | str | None:
        return self._end_time

    @end_time.setter
    def end_time(self, end_time: datetime | str | None):
        if not isinstance(end_time, (type(None), datetime, str)):
            raise TypeError(
                f"end_time is expected ot be an instance of datetime or None. Not {type(end_time)}"
            )
        self._end_time = end_time

    @property
    def title(self) -> str | None:
        return self._title

    @title.setter
    def title(self, title: str | None):
        if isinstance(title, numpy.ndarray):
            # handle diamond use case
            title = str(title)
        elif not isinstance(title, (type(None), str)):
            raise TypeError(
                f"title is expected ot be an instance of str or None. Not {type(title)}"
            )
        self._title = title

    @property
    def instrument(self) -> NXinstrument | None:
        return self._instrument

    @instrument.setter
    def instrument(self, instrument: NXinstrument | None) -> None:
        if not isinstance(instrument, (type(None), NXinstrument)):
            raise TypeError(
                f"instrument is expected ot be an instance of {NXinstrument} or None. Not {type(instrument)}"
            )
        self._instrument = instrument

    @property
    def sample(self) -> NXsample | None:
        return self._sample

    @sample.setter
    def sample(self, sample: NXsample | None):
        if not isinstance(sample, (type(None), NXsample)):
            raise TypeError(
                f"sample is expected ot be an instance of {NXsample} or None. Not {type(sample)}"
            )
        self._sample = sample

    @property
    def control(self) -> NXmonitor | None:
        return self._control

    @control.setter
    def control(self, control: NXmonitor | None) -> None:
        if not isinstance(control, (type(None), NXmonitor)):
            raise TypeError(
                f"control is expected ot be an instance of {NXmonitor} or None. Not {type(control)}"
            )
        self._control = control

    @property
    def energy(self) -> pint.Quantity | None:
        return self._energy

    @energy.setter
    @check_dimensionality(expected_dimension="[energy]")
    def energy(self, energy: pint.Quantity | None) -> None:
        if energy is None:
            self._energy = None
        elif isinstance(energy, pint.Quantity):
            self._energy = energy.to(_ureg.keV)
        else:
            raise TypeError(
                f"energy is expected to be a pint.Quantity or None. Not {type(energy)}"
            )

    @property
    def group_size(self) -> int | None:
        return self._group_size

    @group_size.setter
    def group_size(self, group_size: int | None):
        if not (
            isinstance(group_size, (type(None), int))
            or (numpy.isscalar(group_size) and not isinstance(group_size, (str, bytes)))
        ):
            raise TypeError(
                f"group_size is expected ot be None or a scalar. Not {type(group_size)}"
            )
        self._group_size = group_size

    @property
    def bliss_original_files(self) -> tuple | None:
        return self._bliss_original_files

    @bliss_original_files.setter
    def bliss_original_files(self, files: tuple | numpy.ndarray | None):
        if isinstance(files, numpy.ndarray):
            files = tuple(files)
        if not isinstance(files, (type(None), tuple)):
            raise TypeError(
                f"files is expected to be None or a tuple. {type(files)} provided instead"
            )
        self._bliss_original_files = files

    @docstring(NXobject)
    def to_nx_dict(
        self,
        nexus_path_version: float | None = None,
        data_path: str | None = None,
    ) -> dict:
        if data_path is None:
            data_path = ""

        nexus_paths = get_nexus_paths(nexus_path_version)
        nx_dict = {}

        if self.sample is not None:
            nx_dict.update(
                self.sample.to_nx_dict(nexus_path_version=nexus_path_version)
            )
        else:
            _logger.info("no sample found. Won't be saved")

        if self.instrument is not None:
            nx_dict.update(
                self.instrument.to_nx_dict(nexus_path_version=nexus_path_version)
            )
        else:
            _logger.info("no instrument found. Won't be saved")

        if self.control is not None:
            nx_dict.update(
                self.control.to_nx_dict(nexus_path_version=nexus_path_version)
            )
        else:
            _logger.info("no control found. Won't be saved")

        if self.start_time is not None:
            path_start_time = f"{self.path}/{nexus_paths.START_TIME_PATH}"
            if isinstance(self.start_time, datetime):
                start_time = self.start_time.isoformat()
            else:
                start_time = self.start_time
            nx_dict[path_start_time] = start_time
        if self.end_time is not None:
            path_end_time = f"{self.path}/{nexus_paths.END_TIME_PATH}"
            if isinstance(self.end_time, datetime):
                end_time = self.end_time.isoformat()
            else:
                end_time = self.end_time
            nx_dict[path_end_time] = end_time
        if self.group_size is not None:
            path_grp_size = f"{self.path}/{nexus_paths.GRP_SIZE_ATTR}"
            nx_dict[path_grp_size] = self.group_size
        if self.energy is not None:
            path_energy = f"{self.path}/{nexus_paths.ENERGY_PATH}"

            nx_dict[path_energy] = self.energy.magnitude
            nx_dict["@".join([path_energy, "units"])] = f"{self.energy.units:~}"
            path_beam = f"{self.path}/{nexus_paths.BEAM_PATH}"
            nx_dict["@".join([path_beam, "NX_class"])] = "NXbeam"

            if nexus_paths.VERSION > 1.0:
                nx_dict[f">/{self.path}/beam/incident_energy"] = (
                    f"/{data_path}/{self.path}/{nexus_paths.ENERGY_PATH}"
                )
        if self.title is not None:
            path_title = f"{self.path}/{nexus_paths.NAME_PATH}"
            nx_dict[path_title] = self.title

        if self.bliss_original_files is not None:
            nx_dict[f"/{self.path}/bliss_original_files"] = self.bliss_original_files

        # create data group from symbolic links

        if self.instrument.detector.image_key is not None:
            nx_dict[f">/{self.path}/data/image_key"] = (
                f"/{data_path}/{self.instrument.detector.path}/{nexus_paths.nx_detector_paths.IMAGE_KEY}"
            )
            nx_dict[f">/{self.path}/data/image_key_control"] = (
                f"/{data_path}/{self.instrument.detector.path}/{nexus_paths.nx_detector_paths.IMAGE_KEY_CONTROL}"
            )
        if self.instrument.detector.data is not None:
            nx_dict[f">/{self.path}/data/data"] = (
                f"/{data_path}/{self.instrument.detector.path}/{nexus_paths.nx_detector_paths.DATA}"
            )
            nx_dict[f"/{self.path}/data@NX_class"] = "NXdata"
            nx_dict[f"/{self.path}/data@signal"] = "data"
            nx_dict[f"/{self.path}@default"] = "data"
            nx_dict[f"{self.path}/data@SILX_style/axis_scale_types"] = [
                "linear",
                "linear",
            ]
        if self.sample.rotation_angle is not None:
            nx_dict[f">/{self.path}/data/rotation_angle"] = (
                f"/{data_path}/{self.sample.path}/{nexus_paths.nx_sample_paths.ROTATION_ANGLE}"
            )

        if nx_dict != {}:
            nx_dict[f"{self.path}@NX_class"] = "NXentry"
            nx_dict[f"{self.path}@definition"] = "NXtomo"
            nx_dict[f"{self.path}/definition"] = "NXtomo"
            nx_dict[f"{self.path}@version"] = nexus_paths.VERSION

            if self._coordinate_system is not None:
                nx_dict[f"{self.path}@NeXus_Coordinate_System"] = (
                    self._coordinate_system.value
                )

        return nx_dict

    def detector_data_is_defined_by_url(self) -> bool:
        return self._detector_data_is_defined_by_type(DataUrl)

    def detector_data_is_defined_by_virtual_source(self) -> bool:
        return self._detector_data_is_defined_by_type(h5py.VirtualSource)

    def _detector_data_is_defined_by_type(self, type_):
        return (
            self.instrument is not None
            and self.instrument.detector is not None
            and self.instrument.detector.data is not None
            and isinstance(self.instrument.detector.data, (str, tuple))
            and isinstance(self.instrument.detector.data[0], type_)
        )

    def load(
        self, file_path: str, data_path: str, detector_data_as="as_data_url"
    ) -> NXobject:
        """
        Load NXtomo instance from file_path and data_path

        :param file_path: hdf5 file path containing the NXtomo
        :param data_path: location of the NXtomo
        :param detector_data_as: how to load detector data. Can be:
                                     * "as_virtual_source": load it as h5py's VirtualGroup
                                     * "as_data_url": load it as silx's DataUrl
                                     * "as_numpy_array": load them as a numpy array (warning: can be memory consuming since all the data will be loaded)
        """
        possible_as_values = ("as_virtual_source", "as_data_url", "as_numpy_array")
        if detector_data_as not in possible_as_values:
            raise ValueError(
                f"detector_data_as is expected to be in {possible_as_values} and not {detector_data_as}"
            )

        if not os.path.exists(file_path):
            raise IOError(f"{file_path} does not exists")
        with hdf5_open(file_path) as h5f:
            if data_path not in h5f:
                raise ValueError(f"{data_path} cannot be find in {file_path}")
            root_node = h5f[data_path]

            if "version" in root_node.attrs:
                nexus_version = root_node.attrs["version"]
            else:
                _logger.warning(
                    f"Unable to find nexus version associated with {data_path}@{file_path}"
                )
                nexus_version = LATEST_NXTOMO_VERSION

            coordinate_system = root_node.attrs.get("NeXus_Coordinate_System", None)
            if coordinate_system is not None:
                coordinate_system = CoordinateSystem(coordinate_system)
            self._coordinate_system = coordinate_system

        nexus_paths = get_nexus_paths(nexus_version)
        self.energy = get_quantity(
            file_path=file_path,
            data_path="/".join([data_path, nexus_paths.ENERGY_PATH]),
            default_unit=_ureg.keV,
        )
        start_time = get_data(
            file_path=file_path,
            data_path="/".join([data_path, nexus_paths.START_TIME_PATH]),
        )
        try:
            start_time = datetime.fromisoformat(start_time)
        except Exception:
            start_time = str(start_time) if start_time is not None else None
        self.start_time = start_time

        end_time = get_data(
            file_path=file_path,
            data_path="/".join([data_path, nexus_paths.END_TIME_PATH]),
        )
        try:
            end_time = datetime.fromisoformat(end_time)
        except Exception:
            end_time = str(end_time) if end_time is not None else None
        self.end_time = end_time

        self.bliss_original_files = get_data(
            file_path=file_path,
            data_path="/".join([data_path, "bliss_original_files"]),
        )

        self.title = get_data(
            file_path=file_path, data_path="/".join([data_path, nexus_paths.NAME_PATH])
        )

        self.sample._load(
            file_path, "/".join([data_path, "sample"]), nexus_version=nexus_version
        )
        self.instrument._load(
            file_path,
            "/".join([data_path, "instrument"]),
            nexus_version=nexus_version,
            detector_data_as=detector_data_as,
        )
        self.control._load(
            file_path, "/".join([data_path, "control"]), nexus_version=nexus_version
        )
        return self

    @staticmethod
    def check_consistency(nx_tomo, raises_error: bool = False):
        """
        Ensure some key datasets have the expected number of values.

        :param NXtomo nx_tomo: NXtomo to check
        :param raises_error: if True, raise ValueError when some incoherent number of values is encountered (if missing will drop a warning only).
            if False, only warnings will be issued.
        """
        if not isinstance(nx_tomo, NXtomo):
            raise TypeError(
                f"nx_tomo is expected to be an instance of {NXtomo}. {type(nx_tomo)} provided"
            )
        if nx_tomo.sample is not None:
            n_rotation_angle = (
                len(nx_tomo.sample.rotation_angle)
                if nx_tomo.sample.rotation_angle is not None
                else None
            )
            n_x_trans = (
                len(nx_tomo.sample.x_translation)
                if nx_tomo.sample.x_translation is not None
                else None
            )
            n_y_trans = (
                len(nx_tomo.sample.y_translation)
                if nx_tomo.sample.y_translation is not None
                else None
            )
            n_z_trans = (
                len(nx_tomo.sample.z_translation)
                if nx_tomo.sample.z_translation is not None
                else None
            )
        else:
            n_rotation_angle = None
            n_x_trans = None
            n_y_trans = None
            n_z_trans = None

        if nx_tomo.instrument is not None and nx_tomo.instrument.detector is not None:
            frames = (
                nx_tomo.instrument.detector.data
                if nx_tomo.instrument.detector.data is not None
                else None
            )
            n_frames = len(frames) if frames is not None else None
            image_keys = (
                nx_tomo.instrument.detector.image_key_control
                if nx_tomo.instrument.detector.image_key_control is not None
                else None
            )
            n_image_key = len(image_keys) if image_keys is not None else None
            n_count_time = (
                len(nx_tomo.instrument.detector.count_time)
                if nx_tomo.instrument.detector.count_time is not None
                else None
            )
        else:
            frames = None
            n_frames = None
            n_image_key = None
            image_keys = None
            n_count_time = None

        n_expected_frames = max(
            (n_rotation_angle or 0),
            (n_frames or 0),
            (n_image_key or 0),
            (n_x_trans or 0),
            (n_y_trans or 0),
            (n_z_trans or 0),
        )

        def check(nb_values, info):
            if nb_values is None:
                _logger.warning(f"{info} not defined")
            elif nb_values != n_expected_frames:
                mess = (
                    f"{info} has {nb_values} values when {n_expected_frames} expected"
                )
                if raises_error:
                    raise ValueError(mess)
                else:
                    _logger.warning(mess)

        check(n_rotation_angle, f"{nx_tomo.node_name}.sample.rotation_angle")
        check(n_x_trans, f"{nx_tomo.node_name}.sample.x_translation")
        check(n_y_trans, f"{nx_tomo.node_name}.sample.y_translation")
        check(n_z_trans, f"{nx_tomo.node_name}.sample.z_translation")
        check(n_frames, f"{nx_tomo.node_name}.instrument.detector.data")
        check(n_image_key, f"{nx_tomo.node_name}.instrument.detector.image_key_control")
        check(n_count_time, f"{nx_tomo.node_name}.instrument.detector.count_time")

        tomo_n = (
            nx_tomo.instrument.detector.tomo_n
            if (
                nx_tomo.instrument is not None
                and nx_tomo.instrument.detector is not None
            )
            else None
        )
        if tomo_n is not None and frames is not None:
            n_projection = len(frames[image_keys == ImageKey.PROJECTION.value])
            if n_projection != tomo_n:
                mess = f"incoherent number of projections found ({n_projection}) compared to tomo_n ({tomo_n})"
                if raises_error:
                    raise ValueError(mess)
                else:
                    _logger.warning(mess)

    @staticmethod
    @docstring(NXobject)
    def concatenate(nx_objects: tuple):
        """
        Concatenate a tuple of NXobject instances into a single NXobject.

        :param nx_objects:
        :return: NXtomo instance which is the concatenation of the nx_objects
        """
        nx_objects = tuple(filter(partial(is_not, None), nx_objects))
        # filter None obj
        if len(nx_objects) == 0:
            return None
        # warning: later we make the assumption that nx_objects contains at least one element
        for nx_obj in nx_objects:
            if not isinstance(nx_obj, NXtomo):
                raise TypeError("Cannot concatenate non NXtomo object")

        nx_tomo = NXtomo()

        # check object concatenation can be handled
        def get_energy() -> pint.Quantity | None:
            """
            Determines the expected energy value from a list of NXobjects.
            Ensures energy values are consistent across NXobjects.
            """
            nxtomos_with_energy = filter(
                lambda energy: energy is not None,
                nx_objects,
            )
            try:
                first_nx_tomo_with_energy = next(nxtomos_with_energy)
            except StopIteration:
                return None
            else:
                return first_nx_tomo_with_energy.energy

        nx_tomo.energy = get_energy()

        _logger.info(f"title {nx_objects[0].title} will be picked")
        nx_tomo.title = nx_objects[0].title
        start_times = tuple(
            filter(
                lambda x: x is not None, [nx_obj.start_time for nx_obj in nx_objects]
            )
        )
        end_times = tuple(
            filter(lambda x: x is not None, [nx_obj.end_time for nx_obj in nx_objects])
        )

        nx_tomo.start_time = min(start_times) if len(start_times) > 0 else None
        nx_tomo.end_time = max(end_times) if len(end_times) > 0 else None

        nx_tomo.sample = NXsample.concatenate(
            tuple([nx_obj.sample for nx_obj in nx_objects])
        )
        nx_tomo.sample.parent = nx_tomo

        nx_tomo.instrument = NXinstrument.concatenate(
            tuple([nx_obj.instrument for nx_obj in nx_objects]),
        )
        nx_tomo.instrument.parent = nx_tomo

        nx_tomo.control = NXmonitor.concatenate(
            tuple([nx_obj.control for nx_obj in nx_objects]),
        )
        nx_tomo.control.parent = nx_tomo

        bliss_original_files = set()
        bof_only_none = True
        for nx_obj in nx_objects:
            if nx_obj.bliss_original_files is not None:
                # current behavior of 'bliss_original_files' is that if there is no information (None) then we won't
                # save it to the file as this is a pure 'esrf' information. Else if it is there (even if empty) we save it
                bof_only_none = False
                bliss_original_files.update(nx_obj.bliss_original_files)

        bliss_original_files = tuple(
            sorted(bliss_original_files)
        )  # it is more convenient ot have it sorted - else sorted along obj id
        nx_tomo.bliss_original_files = None if bof_only_none else bliss_original_files

        return nx_tomo

    def check_can_select_from_rotation_angle(self):
        if (
            self.sample is None
            or self.sample.rotation_angle is None
            or len(self.sample.rotation_angle) == 0
        ):
            raise ValueError(
                "No information on rotation angle found. Unable to do a selection based on angles"
            )
        if self.instrument is None or self.instrument.detector is None:
            raise ValueError(
                "No detector found. Unable to do a selection based on angles"
            )

    @docstring(NXobject)
    def save(
        self,
        file_path: str,
        data_path: str,
        nexus_path_version: float | None = None,
        overwrite: bool = False,
    ) -> None:
        # Note: we overwrite save function for NXtomo in order to force 'data_path' to be provided.
        # Else we get both name and data_path and increase complexity to determine
        # the fiinal location
        super().save(
            file_path=file_path,
            data_path=data_path,
            nexus_path_version=nexus_path_version,
            overwrite=overwrite,
        )

    def sub_select_selection_from_angle_range(
        nx_tomo, start_angle: float, stop_angle: float, copy=True
    ):
        """
        Create an NXtomo like `nx_tomo` but update `image_key_control` to INVALID for
        all projections that do not fulfill the condition: start_angle < rotation_angle < stop_angle.

        Note: Darks and flat fields will not be affected by this sub-selection.

        :param start_angle: Left bound for selection (float, in degrees).
        :param stop_angle: Right bound for selection (float, in degrees).
        :param copy: If True, return a copy of nx_tomo; otherwise, modify nx_tomo in place.
        """
        nx_tomo.check_can_select_from_rotation_angle()
        if copy:
            res = deepcopy(nx_tomo)
        else:
            res = nx_tomo

        angles = res.sample.rotation_angle.magnitude

        mask = numpy.logical_and(
            res.instrument.detector.image_key_control == ImageKey.PROJECTION,
            numpy.logical_or(angles < start_angle, angles > stop_angle),
        )
        res.instrument.detector.image_key_control[mask] = ImageKey.INVALID
        return res

    @staticmethod
    def sub_select_from_angle_offset(
        nx_tomo,
        start_angle_offset: float,
        angle_interval: float | None,
        shift_angles: bool,
        copy=True,
    ):
        """
        Get a sub-selection of NXtomo projections that start with a `start_angle_offset` and
        cover `angle_interval`.

        Note: Darks and flat fields will not be affected by this sub-selection.

        :param start_angle_offset: Offset for selection (float, in degrees).
                                The offset is always relative to the first projection angle.
        :param angle_interval: Interval covered by the selection (float, in degrees).
                            If None, selects until the end.
        :param shift_angles: If True, shift angles by `-start_angle_offset` after selection.
        :param copy: If True, return a copy of nx_tomo; otherwise, modify nx_tomo in place.
        """
        nx_tomo.check_can_select_from_rotation_angle()

        if copy:
            res = deepcopy(nx_tomo)
        else:
            res = nx_tomo

        if shift_angles:
            # for the shift we shift all the projection angle. Simpler
            mask_shift = (
                res.instrument.detector.image_key_control == ImageKey.PROJECTION
            )

        # Extract angles as float values
        projection_angles = res.sample.rotation_angle.magnitude[
            res.instrument.detector.image_key_control == ImageKey.PROJECTION
        ]

        if angle_interval is None:
            # Compute full available interval
            angle_interval = abs(
                projection_angles.max() - projection_angles.min()
            ) + abs(start_angle_offset)

        # Determine start and stop angles as floats
        if len(projection_angles) < 2 or projection_angles[1] > projection_angles[0]:
            # rotate with positive angles
            start_angle = projection_angles[0] + start_angle_offset
            stop_angle = start_angle + angle_interval
        else:
            # rotate with negative angles
            start_angle = projection_angles[0] + start_angle_offset
            stop_angle = start_angle - angle_interval

        NXtomo.sub_select_selection_from_angle_range(
            res,
            start_angle=float(start_angle),
            stop_angle=float(stop_angle),
            copy=False,
        )

        if shift_angles:
            angles = res.sample.rotation_angle.magnitude
            angles[mask_shift] -= start_angle_offset
            res.sample.rotation_angle = angles * _ureg.degree

        return res

    @staticmethod
    def clamp_angles(nx_tomo, angle_range, offset=0, copy=True, image_keys=None):
        if copy:
            res = deepcopy(nx_tomo)
        else:
            res = nx_tomo

        if image_keys is None:
            image_keys = ImageKey.values()

        mask_shift = numpy.logical_or(
            *(
                [
                    res.instrument.detector.image_key_control == ImageKey(image_key)
                    for image_key in image_keys
                ]
            )
        )

        angles = res.sample.rotation_angle.magnitude
        angles[mask_shift] -= offset
        angles[mask_shift] = angles[mask_shift] % angle_range
        res.sample.rotation_angle = angles * _ureg.degree
        return res

    @staticmethod
    def get_valid_entries(file_path: str) -> tuple:
        """
        Return the list of 'NXtomo' entries at the root level.

        :param file_path:
        :return: list of valid NXtomo nodes (ordered alphabetically)

        .. note: entries are sorted to ensure consistency
        """
        if not os.path.isfile(file_path):
            raise ValueError("given file path should be a file")

        def browse_group(group):
            res_buf = []
            for entry_alias in group.keys():
                entry = group.get(entry_alias)
                if isinstance(entry, h5py.Group):
                    if NXtomo.node_is_nxtomo(entry):
                        res_buf.append(entry.name)
                    else:
                        res_buf.extend(browse_group(entry))
            return res_buf

        with hdf5_open(file_path) as h5f:
            res = browse_group(h5f)
        res.sort()
        return tuple(res)

    @staticmethod
    def node_is_nxtomo(node: h5py.Group) -> bool:
        """Check whether the given h5py node is an NXtomo node."""
        if "NX_class" in node.attrs or "NXclass" in node.attrs:
            _logger.debug(f"{node.name} is recognized as a nx class.")
        else:
            _logger.debug(f"{node.name} isn't recognized as a nx class.")
            return False
        if "definition" in node.attrs and node.attrs["definition"].lower() == "nxtomo":
            _logger.debug(f"{node.name} is recognized as an NXtomo class.")
            return True
        elif (
            "instrument" in node
            and "NX_class" in node["instrument"].attrs
            and node["instrument"].attrs["NX_class"]
            in (
                "NXinstrument",
                b"NXinstrument",
            )  # b"NXinstrument" is needed for Diamond compatibility
        ):
            return "detector" in node["instrument"]
        return False


def copy_nxtomo_file(
    input_file: str,
    output_file: str,
    entries: tuple | None,
    overwrite: bool = False,
    vds_resolution="update",
):
    """
    Copy one or several NXtomo entries from one file to another (solving relative links).

    :param input_file: NeXus file from which NXtomo entries have to be copied.
    :param output_file: output file.
    :param entries: entries to be copied. If set to None then all entries will be copied.
    :param overwrite: overwrite data path if already exists.
    :param vds_resolution: How to solve virtual datasets. Options are:
        * update: update Virtual source (relative) paths according to the new location of the file.
        * remove: replace the virtual data source by copying directly the resulting dataset. Warning: in this case all the dataset will be loaded in memory.

        In the future another option could be:
        * embed: copy all VDS to new datasets in the output file 'as they are' (avoid to load all the data in memory).

    """
    input_file = os.path.abspath(input_file)
    output_file = os.path.abspath(output_file)
    if input_file == output_file:
        raise ValueError("input file and output file are the same")

    if entries is None:
        entries = NXtomo.get_valid_entries(file_path=input_file)
        if len(entries) == 0:
            _logger.warning(f"no valid entries for {input_file}")

    for entry in entries:
        if vds_resolution == "remove":
            detector_data_as = "as_numpy_array"
        elif vds_resolution == "update":
            detector_data_as = "as_data_url"
        else:
            raise ValueError(
                f"Unexpected value for 'vds_resolution': {vds_resolution}. Valid values are 'remove' and 'update'"
            )

        nx_tomo = NXtomo().load(input_file, entry, detector_data_as=detector_data_as)
        nx_tomo.save(output_file, entry, overwrite=overwrite)
